import soundcloud, requests, os, re, sys
import socket, json

from tqdm import tqdm
from mutagen.mp3 import MP3
from mutagen.mp4 import MP4, MP4Cover
from mutagen.flac import FLAC
from mutagen.id3 import ID3, TIT2, TPE1, TCON, TDRC, APIC
from requests.packages.urllib3.util.retry import Retry
from requests.adapters import HTTPAdapter
from halo import Halo

def set_api_key(key):
    api_key_config = "client_id='{}'".format(key)
    with open('./config.py', 'w') as config_file:
        config_file.write(api_key_config)

def validate_name(name):
        return re.sub('[\\/:*"?<>|]', "_", name)
        
def get_filename(metadata):
    return validate_name("{}-{}.{}".format(metadata["artist"], metadata["title"], metadata["format"]))

def does_file_exist(filename, actual_size=None):
    return os.path.isfile(filename) and (not actual_size or os.path.getsize(filename) >= float(actual_size))

def download_file(session, filename, url, params={}, silent=False):
    if not silent:
        spinner = Halo(text='Connecting to stream...')
        spinner.start()
    response = session.get(url, stream=True, params=params)
    if not silent: spinner.stop()
    file_size = float(response.headers['content-length']) if 'content-length' in response.headers else 0
    if does_file_exist(filename, file_size):
        if not silent: print("{} already exists, skipping\n".format(filename))
        return filename
    if not silent: print("Saving as: {}".format(filename))
    if not silent: print("File Size: {0:.2f}".format(file_size/(1000**2)))
    with open(filename, 'wb') as file:
        for chunk in tqdm(response.iter_content(chunk_size=1024), total=file_size/1024 + 1, unit='KB', unit_scale=True, disable=silent):
            if chunk: file.write(chunk)
    if not silent: print("Download complete\n")

def tag_file(filename, metadata):
    image = None
    if metadata['artwork_url']: download_file('artwork.jpg', metadata['artwork_url'], silent=True)
    if os.path.isfile('artwork.jpg'): image = open('artwork.jpg', 'rb').read()
    if filename.endswith('.mp3'):
        audio = MP3(filename, ID3=ID3)
        audio.add_tags()
        if image: audio.tags.add(APIC(encoding=3, mime='image/jpeg', type=3, desc=u'Cover', data=image))
        audio.tags["TIT2"] = TIT2(encoding=3, text=metadata['title'])
        audio.tags["TPE1"] = TPE1(encoding=3, text=metadata['artist'])
        audio.tags["TDRC"] = TDRC(encoding=3, text=metadata['year'])
        audio.tags["TCON"] = TCON(encoding=3, text= metadata['genre'])
        audio.save()
    elif filename.endswith('.flac'):
        audio = FLAC(filename)
        audio.add_tags()
        if image: audio.tags.add(APIC(encoding=3, mime='image/jpeg', type=3, desc=u'Cover', data=image))
        audio.tags['title'] = metadata['title']
        audio.tags['artist'] = metadata['artist']
        audio.tags['year'] = metadata['year']
        audio.tags['genre'] = metadata['genre']
        audio.save()
    elif filename.endswith('.m4a'):
        audio = MP4(filename)
        audio.add_tags()
        covr = []
        if image: covr.append(MP4Cover(image, MP4Cover.FORMAT_JPEG))
        audio.tags['covr'] = covr
        audio.tags['title'] = metadata['title']
        audio.tags['artist'] = metadata['artist']
        audio.tags['year'] = metadata['year']
        audio.tags['genre'] = metadata['genre']
        audio.save()
    if os.path.isfile('artwork.jpg'): os.remove('artwork.jpg')


from downloader.utils import set_api_key

import sys, argparse, os

def main():
    parser = argparse.ArgumentParser()

    group = parser.add_mutually_exclusive_group(required=False)

    group.add_argument('-t', '--top', action='store_true', default=False,
        help='Downloads the top 10 tracks across all genres')
    group.add_argument('-n', '--new', action='store_true', default=False,
        help='Downloads 10 new tracks across all genres')
    group.add_argument('-u', '--url', default=None, type=str, nargs='?',
        help='URL to download tracks from')

    parser.add_argument('--set-api-key', type=str,
        help='sets the soundcloud API key')
    parser.add_argument('-s', '--similar', action='store_true', default=False,
        help='Downloads 10 tracks similar to the track in the URL')
    parser.add_argument('-d', '--dir', default=os.getcwd(), type=str,
        help='Directory to save tracks in. Defaults to current working directory')

    parser.add_argument('-a', '--all', default=False, action='store_true',
        help='Download all tracks (Uploads and likes)')
    parser.add_argument('-l', '--likes', default=False, action='store_true',
        help='Download only liked tracks.')

    parser.add_argument('-e', '--exclude', nargs='+', type=int,
        help='Enter track numbers to exclude.')
    parser.add_argument('-i', '--include', nargs='+', type=int,
        help='Enter track numbers to include')
    parser.add_argument('--limit', type=int,
        help='Maximum number of tracks to download')
    parser.add_argument('-r', '--range', nargs=2, type=int,
        help='Enter range of tracks to download.')

    parser.add_argument('-g', '--genre', nargs='?', type=str, default='all-music',
        help='use with --top to get top tracks from a specific genre')

    args = parser.parse_args()
    
    if args.set_api_key:
        set_api_key(args.set_api_key)
        return

    if args.similar and not args.url:
        print("Track URL is needed to get similar tracks")
        return

    if not any([args.url, args.top, args.new]):
        print("At least one of the following args is required -n/--new -t/--top or -u/--url")
        return                
    
    from downloader import downloader

    d = downloader.SoundcloudDownloader(args)
    if args.include:
        args.include = set(args.include)
    if args.exclude:
        args.exclude = set(args.exclude)

    try:
        d.main()
    except KeyboardInterrupt:
        print("\nExiting")
        sys.exit(0)

if __name__ == '__main__':
    main()


import json, requests
from config import browser_id, client_id
from requests.adapters import HTTPAdapter

class SoundCloudClient(object):
    def __init__(self, args=None):
        self.API_V1 = "https://api.soundcloud.com"
        self.API_V2 = "https://api-v2.soundcloud.com"
        self.browser_id = browser_id
        self.session = requests.Session()
        self.session.mount("http://", adapter = HTTPAdapter(max_retries = 3))
        self.session.mount("https://", adapter = HTTPAdapter(max_retries = 3))

    def get_uploaded_tracks(self, user_id, limit=9999):
        url_params = {
            "client_id": browser_id,
            "limit": limit ,
            "offset": 0
        }
        url = "{}/users/{}/tracks".format(self.API_V2, user_id)
        response = self.session.get(url, params=url_params)
        json_payload = json.loads(response.text)
        tracks = json_payload["collection"]
        return tracks

    def get_liked_tracks(self, user_id, no_of_tracks=10):
        url_params = {
            "client_id": client_id,
            "limit": no_of_tracks,
            "offset": 0
        }
        url = "{}/users/{}/likes".format(self.API_V2, user_id)
        response = self.session.get(url, params=url_params)
        json_payload = json.loads(response.text)
        tracks = filter(lambda x: 'playlist' not in x, json_payload["collection"])
        return list(map(lambda x: x['track'], tracks))

    def get_recommended_tracks(self, track, no_of_tracks=10):
        url_params = {
            "client_id": browser_id,
            "limit": no_of_tracks,
            "offset": 0
        }
        recommended_tracks_url = "{}/tracks/{}/related".format(self.API_V2, track.id)
        r = self.session.get(recommended_tracks_url, params=url_params)
        tracks = json.loads(r.text)["collection"]
        tracks = map(lambda x: x["track"], tracks[:no_of_tracks])
        return list(tracks)

    def get_charted_tracks(self, kind, genre, limit=9999):
        url_params = {
            "limit": limit,
            "genre": "soundcloud:genres:" + genre,
            "kind": kind,
            "client_id": browser_id
        }
        url = "{}/charts".format(self.API_V2)
        response = self.session.get(url, params=url_params)
        json_payload = json.loads(response.text)
        tracks = json_payload["collection"]
        return tracks



import soundcloud, requests, os, re, sys
import socket, json
from soundcloud import resource
from .utils import download_file, tag_file, validate_name, get_filename
from .context import client_id
from requests.adapters import HTTPAdapter
from halo import Halo

class SoundcloudDownloader(object):
    def __init__(self, args=None):
        self.args = args
        self.url = args.url
        self.dirname = args.dir
        self.client = soundcloud.Client(client_id=client_id)
        self.API_V2 = "https://api-v2.soundcloud.com"
        self.API_V1 = "https://api.soundcloud.com"
        self.download_count = 0
        self.session = requests.Session()
        self.session.mount("http://", adapter = HTTPAdapter(max_retries = 3))
        self.session.mount("https://", adapter = HTTPAdapter(max_retries = 3))

    def can_download_track(self, track):
        is_downloadable = track["downloadable"] and "download_url" in track
        is_streamable = track["streamable"]
        has_stream_url = "stream_url" in track
        for transcoding in track["media"]["transcodings"]:
            if transcoding["format"]["protocol"] == "progressive":
                has_stream_url = True
                break
        return is_downloadable or (is_streamable and has_stream_url)

    def get_track_url(self, track):
        if track["downloadable"] and "download_url" in track:
            return ("{}?client_id={}".format(track["download_url"], client_id), track.get("original_format", "mp3"))
        if track["streamable"]:
            if "stream_url" in track:
                return ("{}?client_id={}".format(track["stream_url"], client_id), "mp3")
            for transcoding in track["media"]["transcodings"]:
                if transcoding["format"]["protocol"] == "progressive":
                    r = self.session.get(transcoding["url"], params={"client_id": client_id})
                    return (json.loads(r.text)["url"] , "mp3")
        return (None, None)

    def get_track_metadata(self, track):
        artist = "unknown"
        if "publisher_metadata" in track and track["publisher_metadata"]:
            artist = track["publisher_metadata"].get("artist", "")
        elif "user" in track or not artist:
            artist = track["user"]["username"]
        url, fileFormat = self.get_track_url(track)
        return {
            "title": str(track.get("title", track["id"])),
            "artist": artist,
            "year": str(track.get("release_year", "")),
            "genre": str(track.get("genre", "")),
            "format": fileFormat,
            "download_url": url,
            "artwork_url": track["artwork_url"]
        }

    def get_single_track(self, track):
        if isinstance(track, resource.Resource): track = track.fields()
        metadata = self.get_track_metadata(track)
        filename = get_filename(metadata)
        if metadata['download_url']:
            download_file(self.session, filename, metadata["download_url"])
            try:
                tag_file(filename, metadata)
            except:
                if os.path.isfile("artwork.jpg"): os.remove("artwork.jpg")
            self.download_count += 1
        else:
            print('Cannot download {}'.format(metadata['title']))

    def get_multiple_tracks(self, tracks):
        for _, track in filter(lambda x: self.check_track_number(x[0]), enumerate(tracks)):
            self.get_single_track(track)

    def get_playlist(self, playlist):
        print("{} tracks found in playlist".format(len(playlist.tracks)))
        self.get_multiple_tracks(playlist.tracks)

    def check_track_number(self, index):
        if self.download_count == self.args.limit:
            return False
        if self.args.include and index + 1 in self.args.include:
            return True
        if self.args.exclude and index + 1 in self.args.exclude:
            return False
        if self.args.range:
            if not self.args.range[0] <= index + 1 <= self.args.range[1]:
                return False
        return True

    def get_recommended_tracks(self, track, no_of_tracks=10):
        params = {
            "client_id": client_id,
            "limit": no_of_tracks,
            "offset": 0
        }
        spinner = Halo(text="Fetching tracks similar to {}".format(track.title))
        spinner.start()
        recommended_tracks_url = "{}/tracks/{}/related".format(self.API_V2, track.id)
        r = self.session.get(recommended_tracks_url, params=params)
        spinner.stop()
        tracks = r.json()["collection"]
        print("Found {} similar tracks".format(len(tracks)))
        self.get_multiple_tracks(tracks)

    def get_charted_tracks(self, kind, no_of_tracks=10):
        url_params = {
            "limit": no_of_tracks,
            "genre": "soundcloud:genres:" + self.args.genre,
            "kind": kind,
            "client_id": client_id
        }
        url = "{}/charts".format(self.API_V2)
        tracks = []
        spinner = Halo(text="Fetching {} {} tracks".format(no_of_tracks, kind))
        spinner.start()
        while len(tracks) < no_of_tracks:
            json_payload = self.session.get(url, params=url_params).json()
            tracks += json_payload["collection"]
            tracks = list(filter(lambda track: self.can_download_track(track['track']), tracks))			
            url = json_payload["next_href"]
        spinner.stop()
        tracks = map(lambda x: x["track"], tracks[:no_of_tracks])
        self.get_multiple_tracks(tracks)

    def get_uploaded_tracks(self, user):
        no_of_tracks = self.args.limit if self.args.limit else 9999
        params = {
            "client_id": client_id,
            "limit": no_of_tracks,
            "offset": 0
        }
        tracks = []
        spinner = Halo(text="Fetching uploads")
        spinner.start()
        url = "{}/users/{}/tracks".format(self.API_V2, user.id)
        while url:
            json_payload = self.session.get(url, params=params).json()
            tracks += json_payload["collection"]
            tracks = list(filter(lambda track: self.can_download_track(track['track']), tracks))
            if len(tracks) >= no_of_tracks: break
            url = json_payload.get("next_href", None)
        spinner.stop()
        print("Found {} uploads".format(len(tracks)))
        self.get_multiple_tracks(tracks)

    def get_liked_tracks(self, user):
        no_of_tracks = self.args.limit if self.args.limit else 9999
        params = {
            "client_id": client_id,
            "limit": no_of_tracks,
            "offset": 0
        }
        tracks = []
        spinner = Halo(text="Fetching uploads")
        spinner.start()
        url = "{}/users/{}/likes".format(self.API_V2, user.id)
        while url:
            json_payload = self.session.get(url, params=params).json()
            tracks += list(filter(lambda x: 'playlist' not in x, json_payload["collection"]))
            tracks = list(filter(lambda track: self.can_download_track(track['track']), tracks))			
            if len(tracks) >= no_of_tracks: break
            url = json_payload.get("next_href", None)
        spinner.stop()
        tracks = [track['track'] for track in tracks]
        print("Found {} likes".format(len(tracks)))
        self.get_multiple_tracks(tracks)

    def main(self):
        os.chdir(self.dirname)
        if self.args.top:
            self.get_charted_tracks("top")
            return 
        if self.args.new:
            self.get_charted_tracks("trending")
            return
        spinner = Halo(text="Resolving URL")
        spinner.start()
        params = {
            "url": self.url,
            "client_id": client_id
        }
        url = "{}/resolve".format(self.API_V1)
        res = self.session.get(url, params=params)
        if not res.ok:
            print("Could not get a valid response from the SoundCloud API. Please check the API key")
            return
        res = res.json()
        data = resource.Resource(res)
        spinner.stop()
        if isinstance(data, resource.Resource):
            if data.kind == "user":
                print("User profile found")
                folder = validate_name(data.username)
                if not os.path.isdir(folder): os.mkdir(folder)
                os.chdir(os.path.join(os.getcwd(), folder))
                print("Saving in: " + os.getcwd())
                if self.args.all or self.args.likes:
                    self.get_liked_tracks(data)
                if not self.args.likes:
                    self.get_uploaded_tracks(data)
            elif data.kind == "track":
                    print("Single track found")
                    print("Saving in: " + os.getcwd())
                    if self.args.similar:
                        self.get_recommended_tracks(data)
                    self.get_single_track(data)
            elif data.kind == "playlist":
                print("Single playlist found.")
                folder = validate_name(data.user["username"])
                if not os.path.isdir(folder):
                    os.mkdir(folder)
                os.chdir(os.path.join(os.getcwd(), str(folder)))
                self.get_playlist(data)
        elif isinstance(data, resource.ResourceList):
            if data[0].kind == "playlist":
                print("%d playlists found" % (len(data)))
                for playlist in data:
                    self.get_playlist(playlist)
            elif data[0].kind == "track":
                self.get_multiple_tracks(data)



import sys
import os
sys.path.insert(0, os.path.abspath(os.path.join(os.path.dirname(__file__), '..')))

try:
    from config import client_id
except ImportError:
    print('API key is not set. Please set it using the --set-api-key option')
    sys.exit(0)